#pragma once

#include <vector>
#include <iostream>

#include <llvm/IR/IRBuilder.h>
#include <llvm/IR/Verifier.h>
#include <llvm/Support/TargetSelect.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/JITSymbol.h>
#include <llvm/ExecutionEngine/SectionMemoryManager.h>
#include <llvm/ExecutionEngine/Orc/CompileUtils.h>
#include <llvm/ExecutionEngine/Orc/IRCompileLayer.h>
#include <llvm/ExecutionEngine/Orc/LambdaResolver.h>
#include <llvm/ExecutionEngine/Orc/RTDyldObjectLinkingLayer.h>

#include "DBData.h"


using namespace llvm;

class JitContext {

public:
    using LinkLayer = orc::RTDyldObjectLinkingLayer;
    using Compiler = orc::SimpleCompiler;
    using CompileLayer = orc::IRCompileLayer<LinkLayer, Compiler>;

    // LLVM Components
    Module* jit_module = nullptr;
    LLVMContext context;
    IRBuilder<> builder;
    TargetMachine* target;

    // Compiler infrastructure    
    std::shared_ptr < CompileLayer > compile_layer;
    std::shared_ptr < LinkLayer > link_layer;

    // Types used in LLVM IR
    Type* i64PtrTy;
    Type* i64Ty;
    Type* i1Ty;

    // Code generation entities
    Function* func;
    Value* arr;
    Value* outArr;
    long int (*query_func) ( long int*, long int* );

    // Relation
    Tuple* rel = nullptr;

    JitContext () : builder ( IRBuilder<>(context) ) {
    	// Initialization
    	LLVMInitializeNativeTarget();
    	InitializeNativeTargetAsmPrinter();
    	target = EngineBuilder().selectTarget();
    	
        // generate code for function
        jit_module = new Module("JIT Query", context);
     
        // types used in llvm code
        i64PtrTy = Type::getInt64PtrTy ( context );
        i64Ty    = Type::getInt64Ty ( context );
        i1Ty     = Type::getInt1Ty ( context );
    
        // create function with parameters ( inRelation, outRelation ) and return numOut
        std::vector<Type*> param_type ( 2, i64PtrTy );
        FunctionType* prototype = FunctionType::get ( i64Ty, param_type, false );
        func = Function::Create ( prototype, Function::ExternalLinkage, "jit_func", jit_module );
        std::vector<Value*> args;
        for(auto& arg : func->args())
                args.push_back(&arg);
        arr = args[0];
        outArr = args[1];
       
        // Create Basic Blocks for initialization and main
        BasicBlock* entry = BasicBlock::Create ( context, "entry", func );
        builder.SetInsertPoint ( entry );
    }

    ~JitContext() {
    	delete target;
    }

    /**
      * @brief Adds sample LLVM IR code one tuple '123' to the output.
      */
    void addSampleCode () {
        Value* outArrElemPtr = builder.CreateGEP ( this->outArr, ConstantInt::get ( this->i64Ty, 0 ), "outArrElemPtr" );
        builder.CreateStore ( ConstantInt::get ( this->i64Ty, 123 ), outArrElemPtr );
        builder.CreateRet ( ConstantInt::get ( this->i64Ty, 1 ) ); 
    }
   
    /**
      * @brief Executes query. Note that 'rel' needs to be assigned by the Scan operator 
      */
    long int execute( long int* outArr ) {
        return query_func ( rel, outArr );
    }

    void addOptPasses ( llvm::legacy::PassManagerBase &passes, llvm::legacy::FunctionPassManager &fnPasses, llvm::TargetMachine *machine );

    void addLinkPasses(llvm::legacy::PassManagerBase &passes);

    void optimizeModule(llvm::TargetMachine *machine, llvm::Module *module);

    static JITSymbol dummy_lookup(const std::string& name) {
        return JITSymbol(nullptr);
    }

    void printModule ( Module* m ) {
        std::string moduleIR;
        raw_string_ostream OS ( moduleIR );
        OS << *jit_module;
        OS.flush();
        std::cout << "------------ Generated LLVM IR Code: ------------" << std::endl 
                  << moduleIR 
                  << "-------------------------------------------------" << std::endl;
    }

    bool compile ( bool printLLVMCode=true ) {
        if ( printLLVMCode ) {
            printModule ( jit_module );
        }
   
        // verify 
    	if ( llvm::verifyModule(*jit_module, &llvm::outs()) ) {
            std::cout << "\nExiting JIT Compilation due to errors." << std::endl;
            return false;
        }

        // optimize module
        //optimizeModule ( target, jit_module );
        //printModule ( jit_module );
  
        // compile
        link_layer = std::make_shared<LinkLayer>([]() { return std::make_shared<SectionMemoryManager>(); });
        compile_layer = std::make_shared<CompileLayer> ( *link_layer, Compiler(*target) );
    	auto jit_module_handle = cantFail ( compile_layer->addModule( std::shared_ptr<Module> ( jit_module ), 
    	    orc::createLambdaResolver ( dummy_lookup, dummy_lookup ) ) );
    	
    	// get function pointer
    	JITSymbol symbol = compile_layer->findSymbolIn( jit_module_handle, "jit_func", false);
        if ( !symbol ) {
            std::cout << "Error: JIT function not found" << std::endl;
            return false;
        }
    	query_func = (decltype(query_func))cantFail(symbol.getAddress());
    	return true;
    }
};

