/**
 * @file
 *
 * Implementation of vector-at-a-time execution for relational operators.
 *
 * @author Henning Funke <henning.funke@cs.tu-dortmund.de>
 *
 * (c) 2019 TU Dortmund, DBIS Group
 */
#include <algorithm>
#include "Operators.h"
#include "primitives.h"


void ScanOp::openVec() {
    this->cursor = 0;
    this->oVec = allocateRelation ( vecLen );
}

Relation* ScanOp::nextVec() {
    if ( cursor >= this->tableSize ) return nullptr;
    size_t len = std::min ( this->tableSize - cursor, (size_t)vecLen ); 
    oVec.len = scanLong ( &table[cursor], oVec.r, len );
    cursor += len;
    return &oVec;
}

void ScanOp::closeVec() {
    freeRelation ( this->oVec );
}


void SelectionOp::openVec() {
    child->openVec();
    this->oVec = allocateRelation ( vecLen );
}

Relation* SelectionOp::nextVec() {
    Relation* in = child->nextVec();
    if ( in == nullptr ) return nullptr;
    Relation* outVec = in;
    switch ( this->type ) {
        case PredicateType::EQUALS:
            outVec->len = compareEquals ( in->r, this->compareConstant, outVec->r, in->len );
            break;
        case PredicateType::EQUALS_NOT:
            outVec->len = compareNotEquals ( in->r, this->compareConstant, outVec->r, in->len );
            break;
        case PredicateType::SMALLER:
            outVec->len = compareSmaller ( in->r, this->compareConstant, outVec->r, in->len );
            break;
    }
    return outVec;
}

void SelectionOp::closeVec() {
    child->closeVec();
    freeRelation ( oVec );
}


void AggregationOp::openVec() {
    child->openVec();
    this->oVec = allocateRelation ( 1 );
    this->hasMoreTuples = true;
    oVec.r[0] = 0;
}

Relation* AggregationOp::nextVec() {
    if ( !this->hasMoreTuples ) return nullptr;
    oVec.len = 1;
    Relation* in = child->nextVec();
    while ( in != nullptr ) {
        if ( this->type == AggregationOp::ReduceType::COUNT ) {
            oVec.r[0] += aggCount ( in->r, in->len );
        }
        else if ( this->type == AggregationOp::ReduceType::SUM ) {
            oVec.r[0] += aggSum ( in->r, in->len );
        }
        in = child->nextVec();
    }
    this->hasMoreTuples = false;
    return &oVec;
}

void AggregationOp::closeVec() {
    child->closeVec();
    freeRelation ( this->oVec );
}
