/**
 * @file
 *
 * WeeDB - Execution engine experiments. 
 *
 * @author Henning Funke <henning.funke@cs.tu-dortmund.de>
 *
 * (c) 2019 TU Dortmund, DBIS Group
 */
#include <iostream>
#include <iomanip>
#include <chrono>
#include <unistd.h>
#include <sstream>

#include "CJit.h"
#include "Jit.h"
#include "DBData.h"
#include "Operators.h"
#include "Timer.h"

#ifndef RELATION_LEN
    #define RELATION_LEN 200000000
#endif
#ifndef VEC_LEN
    #define VEC_LEN 2000
#endif

/**
  * @brief Output header line for csv
  */
void csvHeader () {
    std::cout << std::endl << "VEC_LEN, RELATION_LEN, tVolc, tOp, tVec, tJit, tCJit" << std::endl;
}


/**
  * @brief Output timings of different execution models as csv line
  */
void csvStats ( double tVolc, double tOp, double tVec, double tJit, double tCJit ) {
    std::cout << std::fixed;
    std::cout << std::setprecision(1);
    std::cout << VEC_LEN << ", " << RELATION_LEN << ", " << tVolc 
              << ", " << tOp << ", " << tVec << ", " << tJit << ", " << tCJit <<  std::endl;
}


/**
  * @brief Execute query plan given by root with Volcano (Tuple-at-a-time)
  */
double execVolcano ( RelOperator* root ) { 
    Timer tVolc = Timer();
    Relation rel = allocateRelation ( root->getSize() );
    PullDriver::volcano ( root, &rel );
    printRelation ( rel );
    freeRelation ( rel ); 
    return tVolc.get();
}


/**
  * @brief Execute query plan given by root with Operator-at-a-time
  */
double execOperatorAtATime ( RelOperator* root ) { 
    Timer tOp = Timer();
    Relation resultRelation = root->getRelation();
    printRelation ( resultRelation );
    return tOp.get();
}


/**
  * @brief Execute query plan given by root with Vector-at-a-time
  */
double execVectorized ( RelOperator* root ) { 
    Timer tVec = Timer();
    Relation rel = allocateRelation ( root->getSize() );
    PullDriver::vector ( root, &rel );
    printRelation ( rel );
    freeRelation ( rel ); 
    return tVec.get();
}


/**
  * @brief Execute query plan given by root with Just-in-time Compilation
  */
double execJit ( RelOperator* root, bool printLLVMCode = true ) { 
    Timer tJit = Timer();
    Relation outRel = allocateRelation ( RELATION_LEN ); 
    // Prepend MaterializeOp that writes the result
    root = new MaterializeOp ( root );
    Timer tCompile = Timer();
    JitContext ctx;
    
    ctx.addSampleCode(); // replace with "root->produce ( ctx );" for code generation

    bool compileSuccess = ctx.compile ( printLLVMCode );
    std::cout << "compile time LLVM JIT: " << tCompile.get() << "ms" << std::endl;
    if ( compileSuccess ) {
        outRel.len = ctx.execute ( outRel.r );
    } else {
        std::cout << "compile command returned an error code" << std::endl;
    }
    printRelation ( outRel );
    freeRelation ( outRel ); 
    return tJit.get();
}


/**
  * @brief Execute query plan given by root with Just-in-time Compilation.
  */
double execCJit ( RelOperator* root, bool printCode = false ) { 
    // Prepend MaterializeOp that writes the result
    root = new MaterializeOp ( root );
    Timer tCompile = Timer();
    CJitContext ctx(RELATION_LEN);

    ctx.addSampleCode(); // replace with "root->produceC ( ctx );" for code generation

    bool compileSuccess = ctx.compile();
    std::cout << "compile time C JIT: " << tCompile.get() << "ms" << std::endl;
    if ( compileSuccess ) {
        ctx.execute();
    } else {
        std::cout << "compile command returned an error code" << std::endl;
    }
    // get timing from file 
    std::ifstream t("cjitTime.log");
    std::stringstream buffer;
    buffer << t.rdbuf();
    return std::stod( buffer.str() );
}


/**
  * @brief Parse arguments, generate/read relation, build query plan, and execute.
    The different execution models are selected by arguments or by default we execute all.
  */
int main ( int argc, char* argv[] ) {

    // parse arguments
    bool doVol=false, doOp=false, doVec=false, doJit=false, doCJit=false;
    for (int i = 1; i < argc; ++i) {
        std::string arg = argv[i];
        doVol  |= ( arg.compare ( "vol"  ) == 0 );
        doOp   |= ( arg.compare ( "op"   ) == 0 );
        doVec  |= ( arg.compare ( "vec"  ) == 0 );
        doJit  |= ( arg.compare ( "jit"  ) == 0 );
        doCJit |= ( arg.compare ( "cjit" ) == 0 );
    }
    // none is selected, select all
    if ( ! ( doVol || doOp || doVec || doJit || doCJit ) ) {
        doVol = true; doOp = true; doVec = true; doJit = true; doCJit = true;
    }

    // load or generate relation data
    const char* dbFile = "db.dat";
    Relation relation;
    if ( !loadData ( &relation, dbFile, RELATION_LEN ) ) {
        std::cout << "Generating data.." << std::endl;
        genData ( &relation, dbFile, RELATION_LEN );
    }

    // build plan
    RelOperator* root = new AggregationOp ( AggregationOp::SUM,
        new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 77,
            new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 30,
                new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 99,
                    new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 42,
                        new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 11,
                            new ScanOp ( relation.r, RELATION_LEN )
                        )
                    )
                )
            )
        )
    );


/* --  alternative query plans --
    RelOperator* root = new SelectionOp ( SelectionOp::PredicateType::EQUALS, 11,
        new ScanOp ( relation.r, relation.len )
    );
    
    RelOperator* root = new AggregationOp ( AggregationOp::SUM,
        new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 43,
            new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 42,
                new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 11,
                    new SelectionOp ( SelectionOp::PredicateType::EQUALS_NOT, 12,
                        new ScanOp ( relation.r, relation.len )
                    )
                )
            )
        )
    );

    RelOperator* root = new AggregationOp ( AggregationOp::SUM,
        new ScanOp ( relation.r, RELATION_LEN )
    );
*/    

    double tVol=0.0, tOp=0.0, tVec=0.0, tJit=0.0, tCJit=0.0;

    /* set vector length for vector-at-a-time execution */
    RelOperator::vecLen = VEC_LEN;
    if ( doVol )   tVol  = execVolcano ( root );
    if ( doOp )    tOp   = execOperatorAtATime ( root );
    if ( doVec )   tVec  = execVectorized ( root );
    if ( doJit)    tJit  = execJit ( root );
    if ( doCJit)   tCJit = execCJit ( root );
    csvHeader ();
    csvStats ( tVol, tOp, tVec, tJit, tCJit );

    root->deletePlan();
}



















